@@ -26,10 +26,14 @@ __all__ = [
     'assert_fails_with',
     ]
 
+import warnings
 import sys
 
+from fixtures import Fixture
+
 from testtools.compat import StringIO
-from testtools.content import text_content
+from testtools.content import Content, text_content
+from testtools.content_type import UTF8_TEXT
 from testtools.runtest import RunTest, _raise_force_fail_error
 from testtools._deferred import extract_result
 from testtools._spinner import (
@@ -75,8 +79,13 @@ class SynchronousDeferredRunTest(_DeferredRunTest):
         return result
 
 
-def run_with_log_observers(observers, function, *args, **kwargs):
-    """Run 'function' with the given Twisted log observers."""
+def _get_global_publisher_and_observers():
+    """Return ``(log_publisher, observers)``.
+
+    Twisted 15.2.0 changed the logging framework. This method will always
+    return a tuple of the global log publisher and all observers associated
+    with that publisher.
+    """
     if globalLogPublisher is not None:
         # Twisted >= 15.2.0, with the new twisted.logger framework.
         # log.theLogPublisher.observers will only contain legacy observers;
@@ -86,27 +95,145 @@ def run_with_log_observers(observers, function, *args, **kwargs):
         # observers we want to run with via log.theLogPublisher, because
         # _LogObserver may consider old keys and require them to be mapped.
         publisher = globalLogPublisher
-        real_observers = list(publisher._observers)
+        return (publisher, list(publisher._observers))
     else:
         publisher = log.theLogPublisher
-        real_observers = list(publisher.observers)
-    for observer in real_observers:
-        publisher.removeObserver(observer)
-    for observer in observers:
-        log.theLogPublisher.addObserver(observer)
-    try:
-        return function(*args, **kwargs)
-    finally:
-        for observer in observers:
-            log.theLogPublisher.removeObserver(observer)
-        for observer in real_observers:
-            publisher.addObserver(observer)
+        return (publisher, list(publisher.observers))
+
+
+class _NoTwistedLogObservers(Fixture):
+    """Completely but temporarily remove all Twisted log observers."""
+
+    def _setUp(self):
+        publisher, real_observers = _get_global_publisher_and_observers()
+        for observer in reversed(real_observers):
+            publisher.removeObserver(observer)
+            self.addCleanup(publisher.addObserver, observer)
+
+
+class _TwistedLogObservers(Fixture):
+    """Temporarily add Twisted log observers."""
+
+    def __init__(self, observers):
+        super(_TwistedLogObservers, self).__init__()
+        self._observers = observers
+        self._log_publisher = log.theLogPublisher
+
+    def _setUp(self):
+        for observer in self._observers:
+            self._log_publisher.addObserver(observer)
+            self.addCleanup(self._log_publisher.removeObserver, observer)
+
+
+class _ErrorObserver(Fixture):
+    """Capture errors logged while fixture is active."""
+
+    def __init__(self, error_observer):
+        super(_ErrorObserver, self).__init__()
+        self._error_observer = error_observer
+
+    def _setUp(self):
+        self.useFixture(_TwistedLogObservers([self._error_observer.gotEvent]))
+
+    def flush_logged_errors(self, *error_types):
+        """Clear errors of the given types from the logs.
+
+        If no errors provided, clear all errors.
+
+        :return: An iterable of errors removed from the logs.
+        """
+        return self._error_observer.flushErrors(*error_types)
+
+
+class CaptureTwistedLogs(Fixture):
+    """Capture all the Twisted logs and add them as a detail.
+
+    Much of the time, you won't need to use this directly, as
+    :py:class:`AsynchronousDeferredRunTest` captures Twisted logs when the
+    ``store_twisted_logs`` is set to ``True`` (which it is by default).
+
+    However, if you want to do custom processing of Twisted's logs, then this
+    class can be useful.
+
+    For example::
+
+        class TwistedTests(TestCase):
+            run_tests_with(
+                partial(AsynchronousDeferredRunTest, store_twisted_logs=False))
+
+            def setUp(self):
+                super(TwistedTests, self).setUp()
+                twisted_logs = self.useFixture(CaptureTwistedLogs())
+                # ... do something with twisted_logs ...
+    """
+
+    LOG_DETAIL_NAME = 'twisted-log'
+
+    def _setUp(self):
+        logs = StringIO()
+        full_observer = log.FileLogObserver(logs)
+        self.useFixture(_TwistedLogObservers([full_observer.emit]))
+        self.addDetail(self.LOG_DETAIL_NAME,
+                       Content(UTF8_TEXT, lambda: [logs.getvalue()]))
+
+
+def run_with_log_observers(observers, function, *args, **kwargs):
+    """Run 'function' with the given Twisted log observers."""
+    warnings.warn(
+        'run_with_log_observers is deprecated since 1.8.2.',
+        DeprecationWarning, stacklevel=2)
+    with _NoTwistedLogObservers():
+        with _TwistedLogObservers(observers):
+            return function(*args, **kwargs)
 
 
 # Observer of the Twisted log that we install during tests.
+#
+# This is a global so that users can call flush_logged_errors errors in their
+# test cases.
 _log_observer = _LogObserver()
 
 
+# XXX: Should really be in python-fixtures.
+# See https://github.com/testing-cabal/fixtures/pull/22.
+class _CompoundFixture(Fixture):
+    """A fixture that combines many fixtures."""
+
+    def __init__(self, fixtures):
+        super(_CompoundFixture, self).__init__()
+        self._fixtures = fixtures
+
+    def _setUp(self):
+        for fixture in self._fixtures:
+            self.useFixture(fixture)
+
+
+def flush_logged_errors(*error_types):
+    """Flush errors of the given types from the global Twisted log.
+
+    Any errors logged during a test will be bubbled up to the test result,
+    marking the test as erroring. Use this function to declare that logged
+    errors were expected behavior.
+
+    For example::
+
+        try:
+            1/0
+        except ZeroDivisionError:
+            log.err()
+        # Prevent logged ZeroDivisionError from failing the test.
+        flush_logged_errors(ZeroDivisionError)
+
+    :param error_types: A variable argument list of exception types.
+    """
+    # XXX: jml: I would like to deprecate this in favour of
+    # _ErrorObserver.flush_logged_errors so that I can avoid mutable global
+    # state. However, I don't know how to make the correct instance of
+    # _ErrorObserver.flush_logged_errors available to the end user. I also
+    # don't yet have a clear deprecation/migration path.
+    return _log_observer.flushErrors(*error_types)
+
+
 class AsynchronousDeferredRunTest(_DeferredRunTest):
     """Runner for tests that return Deferreds that fire asynchronously.
 
@@ -115,8 +242,9 @@ class AsynchronousDeferredRunTest(_DeferredRunTest):
     """
 
     def __init__(self, case, handlers=None, last_resort=None, reactor=None,
-                 timeout=0.005, debug=False):
-        """Construct an `AsynchronousDeferredRunTest`.
+                 timeout=0.005, debug=False, suppress_twisted_logging=True,
+                 store_twisted_logs=True):
+        """Construct an ``AsynchronousDeferredRunTest``.
 
         Please be sure to always use keyword syntax, not positional, as the
         base class may add arguments in future - and for core code
@@ -136,6 +264,11 @@ class AsynchronousDeferredRunTest(_DeferredRunTest):
         :param debug: Whether or not to enable Twisted's debugging.  Use this
             to get information about unhandled Deferreds and left-over
             DelayedCalls.  Defaults to False.
+        :param bool suppress_twisted_logging: If True, then suppress Twisted's
+            default logging while the test is being run. Defaults to True.
+        :param bool store_twisted_logs: If True, then store the Twisted logs
+            that took place during the run as the 'twisted-log' detail.
+            Defaults to True.
         """
         super(AsynchronousDeferredRunTest, self).__init__(
             case, handlers, last_resort)
@@ -144,9 +277,12 @@ class AsynchronousDeferredRunTest(_DeferredRunTest):
         self._reactor = reactor
         self._timeout = timeout
         self._debug = debug
+        self._suppress_twisted_logging = suppress_twisted_logging
+        self._store_twisted_logs = store_twisted_logs
 
     @classmethod
-    def make_factory(cls, reactor=None, timeout=0.005, debug=False):
+    def make_factory(cls, reactor=None, timeout=0.005, debug=False,
+                     suppress_twisted_logging=True, store_twisted_logs=True):
         """Make a factory that conforms to the RunTest factory interface.
 
         Example::
@@ -162,7 +298,9 @@ class AsynchronousDeferredRunTest(_DeferredRunTest):
         class AsynchronousDeferredRunTestFactory:
             def __call__(self, case, handlers=None, last_resort=None):
                 return cls(
-                    case, handlers, last_resort, reactor, timeout, debug)
+                    case, handlers, last_resort, reactor, timeout, debug,
+                    suppress_twisted_logging, store_twisted_logs,
+                )
         return AsynchronousDeferredRunTestFactory()
 
     @defer.deferredGenerator
@@ -263,24 +401,38 @@ class AsynchronousDeferredRunTest(_DeferredRunTest):
             self._log_user_exception(TimeoutError(self.case, self._timeout))
             return False, []
 
+    def _get_log_fixture(self):
+        """Return the log fixture we're configured to use."""
+        fixtures = []
+        # TODO: Expose these fixtures and deprecate both of these options in
+        # favour of them.
+        if self._suppress_twisted_logging:
+            fixtures.append(_NoTwistedLogObservers())
+        if self._store_twisted_logs:
+            fixtures.append(CaptureTwistedLogs())
+        return _CompoundFixture(fixtures)
+
     def _run_core(self):
-        # Add an observer to trap all logged errors.
+        # XXX: Blatting over the namespace of the test case isn't a nice thing
+        # to do. Find a better way of communicating between runtest and test
+        # case.
         self.case.reactor = self._reactor
-        error_observer = _log_observer
-        full_log = StringIO()
-        full_observer = log.FileLogObserver(full_log)
         spinner = self._make_spinner()
-        successful, unhandled = run_with_log_observers(
-            [error_observer.gotEvent, full_observer.emit],
-            self._blocking_run_deferred, spinner)
 
-        self.case.addDetail(
-            'twisted-log', text_content(full_log.getvalue()))
-
-        logged_errors = error_observer.flushErrors()
-        for logged_error in logged_errors:
-            successful = False
-            self._got_user_failure(logged_error, tb_label='logged-error')
+        # We can't just install these as fixtures on self.case, because we
+        # need the clean up to run even if the test times out.
+        #
+        # See https://bugs.launchpad.net/testtools/+bug/897196.
+        with self._get_log_fixture() as capture_logs:
+            for name, detail in capture_logs.getDetails().items():
+                self.case.addDetail(name, detail)
+            with _ErrorObserver(_log_observer) as error_fixture:
+                successful, unhandled = self._blocking_run_deferred(
+                    spinner)
+            for logged_error in error_fixture.flush_logged_errors():
+                successful = False
+                self._got_user_failure(
+                    logged_error, tb_label='logged-error')
 
         if unhandled:
             successful = False
@@ -364,27 +516,6 @@ def assert_fails_with(d, *exc_types, **kwargs):
     return d.addCallbacks(got_success, got_failure)
 
 
-def flush_logged_errors(*error_types):
-    """Flush errors of the given types from the global Twisted log.
-
-    Any errors logged during a test will be bubbled up to the test result,
-    marking the test as erroring. Use this function to declare that logged
-    errors were expected behavior.
-
-    For example::
-
-        try:
-            1/0
-        except ZeroDivisionError:
-            log.err()
-        # Prevent logged ZeroDivisionError from failing the test.
-        flush_logged_errors(ZeroDivisionError)
-
-    :param error_types: A variable argument list of exception types.
-    """
-    return _log_observer.flushErrors(*error_types)
-
-
 class UncleanReactorError(Exception):
     """Raised when the reactor has junk in it."""
 
